#pragma once

// https://github.com/ryanhaining/cppitertools

#include "iter/enumerate.hpp"
#include "iter/range.hpp"
#include "iter/reversed.hpp"
#include "iter/sorted.hpp"
#include "iter/zip.hpp"

using iter::enumerate;
using iter::range;
using iter::reversed;
using iter::sorted;
using iter::zip;
